import numpy as np
from datetime import datetime
import os

from pymatgen import Composition
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.entries.computed_entries import ComputedEntry

from maggma.builder import Builder

from atomate.utils.utils import get_mongolike

from monty.serialization import loadfn

import traceback, logging

__author__ = "Joseph Montoya <montoyjh@lbl.gov>"

logger = logging.getLogger(__name__)

module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
settings = loadfn(os.path.join(
    module_dir, "settings", "mpworks_conversion.yaml"))

class MPWorksCompatibilityBuilder(Builder):
    def __init__(self, mpworks_tasks, atomate_tasks, query=None,
                 incremental=True, redo_task_ids=True, **kwargs):
        """
        Converts a task collection from the MPWorks schema to
        Atomate tasks.

        Args:
            mpworks_tasks (Store): Store of task documents
            atomate_tasks (Store): Store of elastic properties
            query (dict): dictionary to limit materials to be analyzed
            incremental (bool): whether to operate in incremental mode,
                i.e. to filter by only tasks that have been updated
                after the last last_updated field in the target store
            redo_task_ids (str/bool): whether to redo_task_ids, if
                a string is supplied, this will be used as the new prefix.
            **kwargs (kwargs): further kwargs for maggma builders
        """

        self.mpworks_tasks = mpworks_tasks
        self.atomate_tasks = atomate_tasks
        self.query = query if query is not None else {}
        self.incremental = incremental
        self.builder_start_time = datetime.utcnow()
        self.redo_task_ids = redo_task_ids

        super().__init__(sources=[mpworks_tasks],
                         targets=[atomate_tasks],
                         **kwargs)

    def connect(self):
        self.mpworks_tasks.connect()
        self.atomate_tasks.connect()

    def get_items(self):
        """
        Gets all items to process into materials documents

        Returns:
            generator or list relevant tasks and materials
            to process into materials documents
        """

        self.logger.info("MPWorks Elastic Compatibility Builder Started")

        # Get only successful tasks
        q = dict(self.query)
        q["state"] = "successful"

        # only consider tasks that have been updated since tasks was last updated
        if self.incremental:
            self.logger.info("Ensuring indices on lu_field")
            self.mpworks_tasks.ensure_index(self.mpworks_tasks.lu_field)
            self.atomate_tasks.ensure_index(self.atomate_tasks.lu_field)
            q.update(self.mpworks_tasks.lu_filter(self.atomate_tasks))

        # No cursor timeout should probably be fixed with smaller batch sizes
        tasks_to_convert = self.mpworks_tasks.query(criteria=q, no_cursor_timeout=True)
        count = tasks_to_convert.count()
        self.logger.info("Found {} new/updated tasks to process".format(count))

        # Redo all task ids at the beginning
        if self.redo_task_ids:
            # Get counter for atomate tasks db
            counter = self.atomate_tasks.collection.database.counter
            counter_doc = counter.find_one({'_id': 'taskid'})
            if not counter_doc:
                counter.insert({"_id": "taskid", "c": 1})
                starting_taskid = 1
            else:
                starting_taskid = counter_doc['c']
            counter.find_one_and_update({"_id": "taskid"}, {"$inc": {"c": count}})

        for n, task in enumerate(tasks_to_convert):
            new_task_id = n + starting_taskid
            self.logger.debug("Processing item: {}->{}, {} of {}".format(
                task['task_id'], new_task_id, n, count))
            yield task, new_task_id

    def process_item(self, item):
        """
        Process the MPWorks tasks and materials
        into an Atomate tasks collection

        Args:
            item (dict): an mpworks document

        Returns:
            an Atomate task document
        """
        mpw_doc, new_task_id = item
        atomate_doc = convert_mpworks_to_atomate(mpw_doc)
        atomate_doc['last_updated'] = self.builder_start_time
        if self.redo_task_ids:
            atomate_doc['_mpworks_meta']['task_id'] = atomate_doc.pop("task_id")
            atomate_doc['task_id'] = new_task_id
        return atomate_doc

    def update_targets(self, items):
        """
        Inserts the new tasks into atomate_tasks collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to
                update and the corresponding processed task_ids
        """
        self.logger.info("Updating {} atomate documents".format(len(items)))

        self.atomate_tasks.update(items, key='task_id')

    def finalize(self, cursor):
        self.atomate_tasks.close()
        self.mpworks_tasks.close()


def convert_mpworks_to_atomate(mpworks_doc, update_mpworks=True):
    """
    Function to convert an mpworks document into an atomate
    document, uses schema above and a few custom cases

    Args:
        mpworks_doc (dict): mpworks task document
        update_mpworks (bool): flag to indicate that mpworks schema
            should be updated to final MPWorks version
    """
    if update_mpworks:
        update_mpworks_schema(mpworks_doc)

    atomate_doc = {}
    for key_mpworks, key_atomate in settings['task_conversion_keys'].items():
        val = get_mongolike(mpworks_doc, key_mpworks)
        set_mongolike(atomate_doc, key_atomate, val)

    # Task type
    atomate_doc["task_label"] = settings['task_label_conversions'].get(
        mpworks_doc["task_type"])

    # calculations
    atomate_doc["calcs_reversed"] = mpworks_doc["calculations"][::-1]

    # anonymous formula
    comp = Composition(atomate_doc['composition_reduced'])
    atomate_doc["formula_anonymous"] = comp.anonymized_formula

    # deformation matrix and original_task_id
    if "deformation_matrix" in mpworks_doc:
        # Transpose this b/c of old bug, should verify in doc processing
        defo = mpworks_doc["deformation_matrix"]
        if isinstance(defo, str):
            defo = convert_string_deformation_to_list(defo)
        defo = np.transpose(defo).tolist()
        set_mongolike(atomate_doc, "transmuter.transformations",
                      ["DeformStructureTransformation"])
        set_mongolike(atomate_doc, "transmuter.transformation_params",
                      [{"deformation": defo}])

    return atomate_doc


def set_mongolike(ddict, key, value):
    """
    Function to set key in a dictionary to value
    using a mongolike key

    Args:
        ddict (dict): dictionary for which to set key
        key (string): key containing strings and/or integers separated
            by periods to indicate subfields, e. g. calcs_reversed.0.output
        value: value to be placed into the dictionary corresponding to
            the mongolike key
    """
    lead_key = key.split('.', 1)[0]
    try:
        lead_key = int(lead_key) # for searching array data
        lead_key_is_int = True
    except ValueError:
        lead_key_is_int = False

    if '.' in key:
        remainder = key.split('.', 1)[1]
        if lead_key not in ddict:
            if lead_key_is_int:
                raise ValueError("Error for key {}, One or more keys is an integer,"
                                 "but no list exists at that key, integer keys may "
                                 "only be specified when a list exists at that key.")
            ddict[lead_key] = {}
        set_mongolike(ddict[lead_key], remainder, value)
    else:
        ddict[key] = value


def convert_string_deformation_to_list(string_defo):
    """
    Some of the older documents in the mpworks schema have a string version
    of the deformation matrix, this function fixes those

    Args:
        string_defo (str): string corresponding to the deformation
    """
    string_defo = string_defo.replace("[", "").replace("]", "").split()
    defo = np.array(string_defo, dtype=float).reshape(3, 3)
    return defo.tolist()


def update_mpworks_schema(mpworks_doc):
    """
    Corrects an mpworks document for outdated schema,
    as enumerated below:

    1. Formats input
    2. Stores final energy according to e_wo_entrop
    3. Tests compatibility, which was added to MPWorks in a later
        iteration

    Args:
        mpworks_doc: document to update schema for

    Returns:
        formatted doc
    """
    # Input
    last_calc = mpworks_doc['calculations'][-1]
    xc = last_calc['input']['incar'].get("GGA")
    if xc:
        xc.upper()
    mpworks_doc["input"].update(
        {"is_lasph": last_calc["input"]["incar"].get("LASPH", False),
         "potcar_spec": last_calc["input"].get("potcar_spec"),
         "xc_override": xc,
         "is_hubbard": last_calc["input"]["incar"].get("LDAU", False)})

    # Final energy - this is being changed because of a change in pymatgen
    #   input parsing that uses e_wo_entrop instead of e_fr_energy
    for calc in mpworks_doc['calculations']:
        total_e = calc['output']['ionic_steps'][-1]['electronic_steps']\
            [-1]['e_wo_entrp']
        e_per_atom = total_e / mpworks_doc['nsites']
        calc['output']['final_energy'] = total_e
        calc['output']['final_energy_per_atom'] = e_per_atom

    # Compatibility
    mpc = MaterialsProjectCompatibility("Advanced")
    func = mpworks_doc["pseudo_potential"]["functional"]
    labels = mpworks_doc["pseudo_potential"]["labels"]
    symbols = ["{} {}".format(func, label) for label in labels]
    parameters = {"run_type": mpworks_doc["run_type"],
                  "is_hubbard": mpworks_doc["is_hubbard"],
                  "hubbards": mpworks_doc["hubbards"],
                  "potcar_symbols": symbols}
    entry = ComputedEntry(Composition(mpworks_doc["unit_cell_formula"]),
                          0.0, 0.0, parameters=parameters,
                          entry_id=mpworks_doc["task_id"])
    try:
        mpworks_doc['is_compatible'] = bool(mpc.process_entry(entry))
    except:
        traceback.print_exc()
        logger.warning('ERROR in getting compatibility')
        mpworks_doc['is_compatible'] = None
    return mpworks_doc



from datetime import datetime
from itertools import chain, groupby
import os

from monty.serialization import loadfn
from pymatgen import Structure
from pymatgen.analysis.structure_matcher import StructureMatcher, ElementComparator

from maggma.builder import Builder
from emmet.vasp.task_tagger import task_type
from pydash.objects import get, set_, has

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"

module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
default_mat_settings = os.path.join(module_dir, "settings", "materials_settings.json")


class MaterialsBuilder(Builder):
    def __init__(self,
                 tasks,
                 materials,
                 mat_prefix="mp-",
                 materials_settings=None,
                 query=None,
                 ltol=0.2,
                 stol=0.3,
                 angle_tol=5,
                 separate_mag_orderings=True,
                 **kwargs):
        """
        Creates a materials collection from tasks and tags

        Args:
            tasks (Store): Store of task documents
            materials (Store): Store of materials documents to generate
            mat_prefix (str): prefix for all materials ids
            materials_settings (Path): Path to settings files
            query (dict): dictionary to limit tasks to be analyzed
            ltol (float): StructureMatcher tuning parameter for matching tasks to materials
            stol (float): StructureMatcher tuning parameter for matching tasks to materials
            angle_tol (float): StructureMatcher tuning parameter for matching tasks to materials
            separate_mag_orderings (bool): Separate magnetic orderings into different materials
        """

        self.tasks = tasks
        self.materials_settings = materials_settings if materials_settings else default_mat_settings
        self.materials = materials
        self.mat_prefix = mat_prefix
        self.query = query if query else {}
        self.ltol = ltol
        self.stol = stol
        self.angle_tol = angle_tol
        self.separate_mag_orderings = separate_mag_orderings

        self.__settings = loadfn(self.materials_settings)

        self.allowed_tasks = {t_type for d in self.__settings for t_type in d['quality_score']}

        super().__init__(sources=[tasks], targets=[materials], **kwargs)

    def get_items(self):
        """
        Gets all items to process into materials documents

        Returns:
            generator or list relevant tasks and materials to process into materials documents
        """

        self.logger.info("Materials Builder Started")
        self.logger.info("Allowed Task Types: {}".format(self.allowed_tasks))

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp for update operation
        self.time_stamp = datetime.utcnow()

        # Get all processed tasks:
        q = dict(self.query)
        q["state"] = "successful"

        all_tasks = set(self.tasks.distinct("task_id", q))
        processed_tasks = set(self.materials.distinct("task_ids"))
        to_process_tasks = all_tasks - processed_tasks
        to_process_forms = self.tasks.distinct("formula_pretty", {"task_id": {"$in": list(to_process_tasks)}})
        self.logger.info("Found {} unprocessed tasks".format(len(to_process_tasks)))
        self.logger.info("Found {} unprocessed formulas".format(len(to_process_forms)))

        # Tasks that have been updated since we last viewed them
        update_q = dict(q)
        update_q.update(self.tasks.lu_filter(self.materials))
        updated_forms = self.tasks.distinct("formula_pretty", update_q)
        self.logger.info("Found {} updated systems to proces".format(len(updated_forms)))

        forms_to_update = set(updated_forms) | set(to_process_forms)
        self.logger.info("Processing {} total systems".format(len(forms_to_update)))

        for formula in forms_to_update:
            tasks_q = dict(q)
            tasks_q["formula_pretty"] = formula
            tasks = list(self.tasks.query(criteria=tasks_q))

            yield tasks

    def process_item(self, tasks):
        """
        Process the tasks into a list of materials

        Args:
            tasks [dict] : a list of task docs

        Returns:
            ([dict],list) : a list of new materials docs and a list of task_ids that were processsed
        """

        formula = tasks[0]["formula_pretty"]
        t_ids = [t["task_id"] for t in tasks]
        self.logger.debug("Processing {} : {}".format(formula, t_ids))

        materials = []
        grouped_tasks = self.filter_and_group_tasks(tasks)

        for group in grouped_tasks:
            materials.append(self.make_mat(group))

        self.logger.debug("Produced {} materials for {}".format(len(materials), tasks[0]["formula_pretty"]))

        return materials

    def make_mat(self, task_group):
        """
        Converts a group of tasks into one material
        """
        all_props = list(chain.from_iterable([self.task_to_prop_list(t) for t in task_group]))
        sorted_props = sorted(all_props, key=lambda x: x['materials_key'])
        grouped_props = groupby(sorted_props, lambda x: x['materials_key'])
        best_props = []
        for name, prop in grouped_props:

            sorted_props = sorted(prop, key=lambda x: x['quality_score'], reverse=True)
            best_props.append(sorted_props[0])

        # Add in the provenance for the properties
        origins = [{k: prop[k]
                    for k in ["materials_key", "task_type", "task_id", "last_updated"]}
                   for prop in best_props
                   if prop.get("track", False)]

        task_ids = list(sorted([t["task_id"] for t in task_group], key=lambda x: int(str(x).split("-")[-1])))

        task_types = {t["task_id"]: t["task_type"] for t in all_props}

        mat = {
            "updated_at": datetime.utcnow(),
            "task_ids": task_ids,
            self.materials.key: task_ids[0],
            "origins": origins,
            "task_types": task_types
        }

        for prop in best_props:
            set_(mat, prop["materials_key"], prop["value"])

        return mat

    def filter_and_group_tasks(self, tasks):
        """
        Groups tasks by structure matching
        """

        filtered_tasks = [t for t in tasks if task_type(t['orig_inputs']) in self.allowed_tasks]

        structures = []

        for idx, t in enumerate(filtered_tasks):
            s = Structure.from_dict(t["output"]['structure'])
            s.index = idx
            structures.append(s)

        if self.separate_mag_orderings:
            for structure in structures:
                if has(structure.site_properties, "magmom"):
                    structure.add_spin_by_site(structure.site_properties['magmom'])
                    structure.remove_site_property('magmom')

        sm = StructureMatcher(
            ltol=self.ltol,
            stol=self.stol,
            angle_tol=self.angle_tol,
            primitive_cell=True,
            scale=True,
            attempt_supercell=False,
            allow_subset=False,
            comparator=ElementComparator())

        grouped_structures = sm.group_structures(structures)

        grouped_tasks = [[filtered_tasks[struc.index] for struc in group] for group in grouped_structures]

        return grouped_tasks

    def task_to_prop_list(self, task):
        """
        Converts a task into an list of properties
        """
        t_type = task_type(task['orig_inputs'])
        t_id = task["task_id"]

        # Convert the task doc into a serious of properties in the materials
        # doc with the right document structure
        props = []
        for prop in self.__settings:
            if t_type in prop["quality_score"].keys():
                if has(task, prop["tasks_key"]):
                    props.append({
                        "value": get(task, prop["tasks_key"]),
                        "task_type": t_type,
                        "task_id": t_id,
                        "quality_score": prop["quality_score"][t_type],
                        "track": prop.get("track", False),
                        "last_updated": task["last_updated"],
                        "materials_key": prop["materials_key"]
                    })
                elif not prop.get("optional", False):
                    self.logger.error("Failed getting {} for task: {}".format(prop["tasks_key"], t_id))
        return props

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the corresponding processed task_ids
        """
        items = [i for i in filter(None, chain.from_iterable(items)) if self.valid(i)]
        if len(items) > 0:
            self.logger.info("Updating {} materials".format(len(items)))
            self.materials.update(docs=items)
        else:
            self.logger.info("No items to update")

    def valid(self, doc):
        """
        Determines if the resulting material document is valid
        """
        return "structure" in doc

    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        :return:
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id", unique=True)
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_pretty")
        self.tasks.ensure_index(self.tasks.lu_field)

        # Search index for materials
        self.materials.ensure_index(self.materials.key, unique=True)
        self.materials.ensure_index("task_ids")



from datetime import datetime

from maggma.builder import Builder
from maggma.utils import recursive_update

class AggregateBuilder(Builder):
    """
    Concats multiple collections together

    Uses `lu_field` to get new/updated documents,
    and uses a `key` field to determine which documents to merge together

    """

    def __init__(self, sources, target, key_field, aggregate_mode="Update", query = {},**kwargs):
        self.key_field = key_field
        self.sources = sources
        self.aggregate_mode = aggregate_mode
        self.target = target
        self.kwargs = kwargs
        self.query = query

        super(AggregateBuilder, self).__init__(sources=sources, targets=[target], **kwargs)

    def get_items(self):

        keys_to_update = set()

        for source in self.sources:
            new_q = dict(self.query)
            new_q.update(source.lu_filter(self.targets))
            keys_to_update |= set(source().distinct(self.key_field,new_q))

        for key in keys_to_update:
            d = {}
            for source in self.sources:
                doc = source().find_one({self.key_field: key})
                if doc:
                    if self.aggregate_mode is "Overwrite":
                        for k,v in doc.items():
                            d[k] = v
                    else:
                        recursive_update(d, doc)
                    d.pop(source.lu_field)
            yield d

    def update_targets(self, items):

        bulk = self.target().initialize_ordered_bulk_op()
        for item in items:
            # Don't alter immutable field _id in target.
            item.pop("_id",None)
            # set a new updated field
            item[self.target.lu_field] = datetime.utcnow()
            bulk.find({self.key_field:item[self.key_field]}).upsert().replace_one(item)
        bulk.execute()



import logging
import numpy as np
from datetime import datetime
from itertools import combinations

from pymongo import ASCENDING

from pymatgen import Structure
from pymatgen.analysis.elasticity.tensors import Tensor
from pymatgen.analysis.piezo import PiezoTensor

from maggma.builder import Builder

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


class DielectricBuilder(Builder):
    def __init__(self, materials, dielectric, min_band_gap=0.1, query={}, **kwargs):
        """
        Creates a dielectric collection for materials

        Args:
            materials (Store): Store of materials documents to match to
            dielectric (Store): Store of dielectric properties
            min_band_gap (float): minimum band gap for a material to look for a dielectric calculation to build
            query (dict): dictionary to limit materials to be analyzed
        """

        self.materials = materials
        self.dielectric = dielectric
        self.min_band_gap = min_band_gap
        self.query = query

        super().__init__(sources=[materials],
                         targets=[dielectric],
                         **kwargs)

    def get_items(self):
        """
        Gets all items to process into materials documents

        Returns:
            generator or list relevant tasks and materials to process into materials documents
        """

        self.logger.info("Dielectric Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        q = dict(self.query)
        q.update(self.materials.lu_filter(self.dielectric))
        q["dielectric"] = {"$exists": 1}
        mats = list(self.materials.find(q, {"material_id": 1}))

        self.logger.info("Found {} new materials for dielectric data".format(len(mats)))

        for m in mats:
            mat = self.materials().find_one(m, {"material_id": 1, "dielectric": 1, "piezo": 1, "structure": 1})
            yield mat

    def process_item(self, item):
        """
        Process the tasks and materials into a dielectrics collection

        Args:
            item dict: a dict of material_id, structure, and tasks

        Returns:
            dict: a dieletrics dictionary  
        """

        def poly(matrix):
            diags = np.diagonal(matrix)
            return np.prod(diags) / np.sum(np.prod(comb) for comb in combinations(diags, 2))

        d = {
            "material_id": item["material_id"]
        }

        structure = Structure.from_dict(item["structure"])

        if item.get("dielectric") is not None:
            ionic = Tensor(d["dielectric"]["ionic"])
            static = Tensor(d["dielectric"]["static"])
            total = ionic + static

            d["dielectric"] = {
                "total": total.symmetrized.fit_to_structure(structure).convert_to_ieee(structure),
                "ionic": ionic.symmetrized.fit_to_structure(structure).convert_to_ieee(structure),
                "static": static.symmetrized.fit_to_structure(structure).convert_to_ieee(structure),
                "e_total": poly(total),
                "e_ionic": poly(ionic),
                "e_static": poly(static)
            }

        # Update piezo if non_centrosymmetric
        if item.get("piezo") is not None:
            static = PiezoTensor.from_voigt(np.array(item['piezo']["piezo_tensor"]))
            ionic = PiezoTensor.from_voigt(np.array(item['piezo']["piezo_ionic_tensor"]))
            total = ionic + static

            d["piezo"] = {
                "total": total.symmetrized.fit_to_structure(structure).convert_to_ieee(structure).voigt,
                "ionic": ionic.symmetrized.fit_to_structure(structure).convert_to_ieee(structure).voigt,
                "static": static.symmetrized.fit_to_structure(structure).convert_to_ieee(structure).voigt,
                "e_ij_max": np.max(total.voigt)
            }

            # TODO Add in more analysis: v_max ?
            # TODO: Add in unstable phonon mode analysis of piezoelectric for potentially ferroelectric

        if len(d) > 1:
            return d

        return None

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the corresponding processed task_ids
        """

        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} dielectrics".format(len(items)))
            bulk = self.dielectric().initialize_ordered_bulk_op()

            for m in filter(None, items):
                m[self.dielectric.lu_field] = datetime.utcnow()
                bulk.find({"material_id": m["material_id"]}).upsert().replace_one(m)
            bulk.execute()
        else:
            self.logger.info("No items to update")

    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        :return:
        """
        # Search index for materials
        self.materials().create_index("material_id", unique=True, background=True)
